/*
    Project: TCGINA
    Version: V1.21
    File   : kfdlg.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#define _WIN32_WINNT 0x0500
#include <tchar.h>
#include "kfdlg.h"
#include "pwdlg.h"
#include "misc.h"
#include "secmem.h"
#include "random.h"

#define KF_MAX_ITEMS 200

static void UpdateCaption(PKF_DIALOG_DATA pkfdd);
static void UpdateList(PKF_DIALOG_DATA pkfdd);
static void NewDocument(PKF_DIALOG_DATA pkfdd);

/*------------------------------------------------------------------------------
//  Password functions
//----------------------------------------------------------------------------*/

static void AddRegistryPassword(PKF_DIALOG_DATA pkfdd, HKEY hKey,
  LPCTSTR pszValueName)
{
    DWORD dwType;
    DWORD cbData;

    cbData = KFI_MAX_PASSWORD*sizeof(TCHAR);

    if (RegQueryValueEx(hKey, pszValueName, NULL, &dwType,
      (LPBYTE)pkfdd->pwszTempPassword, &cbData) == ERROR_SUCCESS) {
        if (dwType == REG_SZ) {
            if (WideCharToMultiByte(CP_UTF8, 0, pkfdd->pwszTempPassword,
              cbData/sizeof(TCHAR), pkfdd->pSecData->rgPasswordCache[
              pkfdd->dwCacheIndex].aszPassword, KFI_MAX_PASSWORD, NULL, NULL)) {
                if (++pkfdd->dwCacheIndex == KF_CACHE_SIZE) {
                    pkfdd->dwCacheIndex = 0;
                }
            } else {
                *pkfdd->pSecData->rgPasswordCache[pkfdd->
                  dwCacheIndex].aszPassword = '\0';
            }
        }
    }
}

#define TCGINA_REGKEY_NAME \
  TEXT("SOFTWARE\\TCGINA DLL\\{D47DF546-F16B-490b-A6D8-523603A5594D}")

static void FillPasswordCache(PKF_DIALOG_DATA pkfdd)
{
    HKEY   hKey;

    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, TCGINA_REGKEY_NAME, 0, KEY_READ,
      &hKey) == ERROR_SUCCESS) {
        DWORD dwIndex = 0;

        AddRegistryPassword(pkfdd, hKey, TEXT("KFPassword"));

        for (;;) {
            HKEY     hSubKey;
            TCHAR    szSubKeyName[256];
            DWORD    cchSubKeySize;
            FILETIME ft;

            cchSubKeySize = sizeof(szSubKeyName)/sizeof(TCHAR);
            if (RegEnumKeyEx(hKey, dwIndex++, szSubKeyName, &cchSubKeySize,
              0, NULL, NULL, &ft) != ERROR_SUCCESS) break;
            if (RegOpenKeyEx(hKey, szSubKeyName, 0, KEY_READ, &hSubKey) ==
              ERROR_SUCCESS) {
                AddRegistryPassword(pkfdd, hSubKey, TEXT("KFPassword"));
                RegCloseKey(hSubKey);
            }
        }
        RegCloseKey(hKey);
    }
}

static BOOL TryPassword(PKF_DIALOG_DATA pkfdd, PKF_ENTRY pEntry)
{
    LPSTR     paszPassword;
    size_t    cbPasswordSize;
    PKF_ENTRY pTempEntry;
    BOOL      fUpdateList;
    BOOL      fResult;

    paszPassword = pkfdd->paszPassword;
    if (*paszPassword == '\0') {
        return FALSE;
    }
    fUpdateList = FALSE;
    fResult = FALSE;
    cbPasswordSize = strlen(paszPassword);

    for (pTempEntry = pkfdd->pFirstEntry; pTempEntry != NULL;
      pTempEntry = pTempEntry->pNextEntry) {
        if (pTempEntry->dwType == KFET_UNKNOWN && pTempEntry != pEntry) {
            if (KFI_InitEncryptionData(&pTempEntry->kfiEncryptedData,
              pkfdd->pkfiKeyData, paszPassword, cbPasswordSize) != KFIE_OK) {
                continue;
            }
            if (KFI_Decrypt(pkfdd->pkfiDecryptedData,
              &pTempEntry->kfiEncryptedData, pkfdd->pkfiKeyData) == KFIE_OK) {
                CopyStringFromUnicode(pTempEntry->szDescription,
                  countof(pTempEntry->szDescription),
                  pkfdd->pkfiDecryptedData->wszDescription);
                pTempEntry->dwID = pkfdd->pkfiDecryptedData->dwID;
                pTempEntry->dwType = KFET_ENCRYPTED;
                fUpdateList = TRUE;
            }
        }
    }
    if (pEntry != NULL) {
        if (KFI_InitEncryptionData(&pEntry->kfiEncryptedData,
          pkfdd->pkfiKeyData, paszPassword, cbPasswordSize) == KFIE_OK) {
            if (KFI_Decrypt(pkfdd->pkfiDecryptedData,
              &pEntry->kfiEncryptedData, pkfdd->pkfiKeyData) == KFIE_OK) {
                if (pEntry->dwType == KFET_UNKNOWN) {
                    CopyStringFromUnicode(pEntry->szDescription,
                      countof(pEntry->szDescription),
                      pkfdd->pkfiDecryptedData->wszDescription);
                    pEntry->dwID = pkfdd->pkfiDecryptedData->dwID;
                    pEntry->dwType = KFET_ENCRYPTED;
                    fUpdateList = TRUE;
                }
                fResult = TRUE;
            }
        }
    }
    if (fUpdateList || fResult || pEntry == NULL) {
        unsigned i;
        for (i = 0; i < KF_CACHE_SIZE; i++) {
            if (strcmp(pkfdd->pSecData->rgPasswordCache[i].aszPassword,
              paszPassword) == 0) {
                break;
            }
        }
        if (i == KF_CACHE_SIZE) {
            CopyStringA(pkfdd->pSecData->
              rgPasswordCache[pkfdd->dwCacheIndex].aszPassword,
              KFI_MAX_PASSWORD, paszPassword);
            if (++pkfdd->dwCacheIndex == KF_CACHE_SIZE) {
                pkfdd->dwCacheIndex = 0;
            }
        }
    }
    if (fUpdateList && pEntry != NULL) {
        UpdateList(pkfdd);
    }

    return fResult;
}

static BOOL TryCachedPasswords(PKF_DIALOG_DATA pkfdd, PKF_ENTRY pEntry)
{
    unsigned i;
    LPSTR    paszPassword;
    size_t   cbPasswordSize;

    for (i = 0; i < KF_CACHE_SIZE; i++) {
        paszPassword = pkfdd->pSecData->rgPasswordCache[i].aszPassword;
        if (*paszPassword == '\0') {
            continue;
        }
        cbPasswordSize = strlen(paszPassword);
        if (KFI_InitEncryptionData(&pEntry->kfiEncryptedData,
          pkfdd->pkfiKeyData, paszPassword, cbPasswordSize) == KFIE_OK) {
            if (KFI_Decrypt(pkfdd->pkfiDecryptedData,
              &pEntry->kfiEncryptedData, pkfdd->pkfiKeyData) == KFIE_OK) {
                CopyStringA(pkfdd->paszPassword, KFI_MAX_PASSWORD,
                  paszPassword);
                return TRUE;
            }
        }
    }

    return FALSE;
}

/*------------------------------------------------------------------------------
//  Read file
//----------------------------------------------------------------------------*/

static BOOL OpenKeyFileDialog(PKF_DIALOG_DATA pkfdd, LPTSTR pszFileName)
{
    OPENFILENAME ofn;

    memset(&ofn, '\0', sizeof(ofn));
    ofn.lStructSize = sizeof(ofn);
    ofn.hwndOwner = pkfdd->hwndDlg;
    ofn.lpstrFilter = TEXT("All Files\0*.*\0\0");
    ofn.lpstrFile = pszFileName;
    ofn.nMaxFile = MAX_PATH;
    ofn.lpstrTitle = TEXT("Open Keyfile");
    ofn.Flags = OFN_HIDEREADONLY | OFN_DONTADDTORECENT;
    if (GetOpenFileName(&ofn)) {
        return TRUE;
    } else if (CommDlgExtendedError() == FNERR_INVALIDFILENAME) {
        *pszFileName = TEXT('\0');
        if (GetOpenFileName(&ofn)) {
            return TRUE;
        }
    }

    return FALSE;
}

static BOOL OpenKeyFile(PKF_DIALOG_DATA pkfdd, LPTSTR pszFileName)
{
    PKF_ENTRY     pEntry;
    PKF_ENTRY     pPrevEntry;
    HANDLE        hFile;
    LARGE_INTEGER liFileSize;
    DWORD         dwCount;
    DWORD         cbRead;
    DWORD         dwTemp;

    hFile = CreateFile(pszFileName, GENERIC_READ, FILE_SHARE_READ, NULL,
      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE) {
        return FALSE;
    }
    if (!GetFileSizeEx(hFile, &liFileSize)) {
        CloseHandle(hFile);
        return FALSE;
    }
    if (liFileSize.QuadPart < sizeof(KFI_DATA) ||
      liFileSize.QuadPart > KF_MAX_ITEMS*sizeof(KFI_DATA) ||
      liFileSize.QuadPart % sizeof(KFI_DATA) != 0) {
        CloseHandle(hFile);
        return FALSE;
    }
    dwCount = (DWORD) (liFileSize.QuadPart / sizeof(KFI_DATA));

    NewDocument(pkfdd);

    pPrevEntry = NULL;
    for (dwTemp = 0; dwTemp < dwCount; dwTemp++) {
        if ((pEntry = malloc(sizeof(KF_ENTRY))) == NULL) {
            break;
        }
        if (pPrevEntry) {
            pPrevEntry->pNextEntry = pEntry;
        } else {
            pkfdd->pFirstEntry = pEntry;
        }
        memset(pEntry, '\0', sizeof(*pEntry));
        if (!ReadFile(hFile, &pEntry->kfiEncryptedData, sizeof(KFI_DATA),
          &cbRead, NULL) || cbRead != sizeof(KFI_DATA)) {
            break;
        }
        pEntry->dwType = KFET_UNKNOWN;
        pEntry->dwTag = dwTemp + 1;
        pPrevEntry = pEntry;
    }
    CloseHandle(hFile);
    if (dwTemp < dwCount) {
        NewDocument(pkfdd);
        return FALSE;
    }
    pkfdd->pLastEntry = pPrevEntry;
    pkfdd->dwCount = dwCount;

    SetCursor(pkfdd->hWaitCursor);

    pkfdd->pkfiKeyData->fValid = FALSE;
    for (pEntry = pkfdd->pFirstEntry; pEntry != NULL;
      pEntry = pEntry->pNextEntry) {
        if (KFI_Decrypt(pkfdd->pkfiDecryptedData, &pEntry->kfiEncryptedData,
          pkfdd->pkfiKeyData) != KFIE_OK) {
            continue;
        }
        pEntry->dwType = KFET_PLAINTEXT;
        pEntry->dwID = pkfdd->pkfiDecryptedData->dwID;
        CopyStringFromUnicode(pEntry->szDescription,
          countof(pEntry->szDescription),
          pkfdd->pkfiDecryptedData->wszDescription);
    }
    for (pEntry = pkfdd->pFirstEntry; pEntry != NULL;
      pEntry = pEntry->pNextEntry) {
        if (pEntry->dwType == KFET_UNKNOWN &&
          TryCachedPasswords(pkfdd, pEntry)) {
            pEntry->dwType = KFET_ENCRYPTED;
            pEntry->dwID = pkfdd->pkfiDecryptedData->dwID;
            CopyStringFromUnicode(pEntry->szDescription,
              countof(pEntry->szDescription),
              pkfdd->pkfiDecryptedData->wszDescription);
        }
    }

    UpdateList(pkfdd);
    if (pkfdd->szFileName != pszFileName) {
        CopyString(pkfdd->szFileName, countof(pkfdd->szFileName), pszFileName);
    }
    UpdateCaption(pkfdd);
    pkfdd->fModified = FALSE;
    SetCursor(pkfdd->hArrowCursor);

    return TRUE;
}

/*------------------------------------------------------------------------------
//  Write file
//----------------------------------------------------------------------------*/

static BOOL SaveFileDialog(PKF_DIALOG_DATA pkfdd, LPTSTR pszFileName)
{
    OPENFILENAME ofn;

    memset(&ofn, '\0', sizeof(ofn));
    ofn.lStructSize = sizeof(ofn);
    ofn.hwndOwner = pkfdd->hwndDlg;
    ofn.lpstrFilter = TEXT("All Files\0*.*\0\0");
    ofn.lpstrFile = pszFileName;
    ofn.nMaxFile = MAX_PATH;
    ofn.lpstrTitle = TEXT("Save Keyfile");
    ofn.Flags = OFN_HIDEREADONLY | OFN_DONTADDTORECENT | OFN_OVERWRITEPROMPT;
    if (GetSaveFileName(&ofn)) {
        return TRUE;
    } else if (CommDlgExtendedError() == FNERR_INVALIDFILENAME) {
        *pszFileName = TEXT('\0');
        if (GetSaveFileName(&ofn)) {
            return TRUE;
        }
    }

    return FALSE;
}

static BOOL SaveFile(PKF_DIALOG_DATA pkfdd, LPTSTR pszFileName)
{
    PKF_ENTRY pEntry;
    HANDLE    hFile;
    DWORD     cbWritten;
    BOOL      fResult;

    if (*pkfdd->szFileName == TEXT('\0')) {
        DWORD dwRandomCount = 0;
        DWORD dwEncryptedCount = 0;
        for (pEntry = pkfdd->pFirstEntry; pEntry != NULL;
          pEntry = pEntry->pNextEntry) {
            if (pEntry->dwType == KFET_ENCRYPTED) {
                dwEncryptedCount++;
            } else if (pEntry->dwType == KFET_UNKNOWN) {
                dwRandomCount++;
            }
        }
        if (dwRandomCount < dwEncryptedCount && pkfdd->dwCount < KF_MAX_ITEMS) {
            BOOL fOneSingleItem;
            dwRandomCount = dwEncryptedCount - dwRandomCount;
            fOneSingleItem = dwRandomCount == 1;
            while (dwRandomCount--) {
                SendMessage(pkfdd->hwndDlg, WM_COMMAND, ID_ITEM_NEW_RANDOM, 0);
            }
            MessageBox(pkfdd->hwndDlg, fOneSingleItem ?
              TEXT("One random item has been added for plausible deniability.")
              : TEXT("Random items have been added for plausible deniability."),
              TEXT("Keyfile Editor"), MB_ICONINFORMATION|MB_OK);
        }
    }

    hFile = CreateFile(pszFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
      FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE) {
        return FALSE;
    }

    for (pEntry = pkfdd->pFirstEntry; pEntry != NULL;
      pEntry = pEntry->pNextEntry) {
        if (!WriteFile(hFile, &pEntry->kfiEncryptedData, sizeof(KFI_DATA),
          &cbWritten, NULL) || cbWritten != sizeof(KFI_DATA)) {
            fResult = FALSE;
            goto error;
         }
    }
    fResult = TRUE;
    if (pkfdd->szFileName != pszFileName) {
        CopyString(pkfdd->szFileName, countof(pkfdd->szFileName), pszFileName);
    }
    pkfdd->fModified = FALSE;
    for (pEntry = pkfdd->pFirstEntry; pEntry != NULL;
      pEntry = pEntry->pNextEntry) {
        pEntry->fModified = FALSE;
    }

error:
    CloseHandle(hFile);
    if (fResult) {
        UpdateCaption(pkfdd);
    } else {
        DeleteFile(pszFileName);
    }

    return fResult;
}

static BOOL SaveQuery(PKF_DIALOG_DATA pkfdd)
{
    if (pkfdd->fModified && pkfdd->pFirstEntry != NULL) {
        switch (MessageBox(pkfdd->hwndDlg,
          TEXT("Keyfile has been modified. Save changes?"),
          TEXT("Keyfile Editor"), MB_ICONQUESTION|MB_YESNOCANCEL)) {
        case IDYES: {
            TCHAR szFileName[MAX_PATH];
            CopyString(szFileName, MAX_PATH, pkfdd->szFileName);
            if (SaveFileDialog(pkfdd, szFileName)) {
                if (SaveFile(pkfdd, szFileName)) {
                    return TRUE;
                }
            }
            /* fall through */ }
        default:
            return FALSE;
        case IDNO:
            break;
        }
    }

    return TRUE;
}

/*------------------------------------------------------------------------------
//  Dialog functions
//----------------------------------------------------------------------------*/

#define MAX_LIST_TEXT_LEN (KFI_MAX_DESCRIPTION + 20)

static void UpdateCaption(PKF_DIALOG_DATA pkfdd)
{
    TCHAR  szCaption[MAX_PATH+200];
    LPTSTR pszFileName;

    pszFileName = _tcsrchr(pkfdd->szFileName, TEXT('\\'));
    if (pszFileName == NULL) {
        pszFileName = _tcsrchr(pkfdd->szFileName, TEXT('/'));
        if (pszFileName == NULL) {
            pszFileName = pkfdd->szFileName-1;
        }
    }
    pszFileName++;
    if (*pszFileName != TEXT('\0')) {
        _sntprintf(szCaption, countof(szCaption), TEXT("%s - Keyfile Editor"),
          pszFileName);
    } else {
        _sntprintf(szCaption, countof(szCaption), TEXT("Keyfile Editor"));
    }
    SetWindowText(pkfdd->hwndDlg, szCaption);
}

static void UpdateList(PKF_DIALOG_DATA pkfdd)
{
    TCHAR     szName[MAX_LIST_TEXT_LEN + 1];
    PKF_ENTRY pEntry;
    LRESULT   lCurSel;

    lCurSel = SendMessage(pkfdd->hwndList, LB_GETCURSEL, 0, 0);
    SendMessage(pkfdd->hwndList, LB_RESETCONTENT, 0, 0);
    if (pkfdd->dwCount) {
        for (pEntry = pkfdd->pFirstEntry; pEntry != NULL;
          pEntry = pEntry->pNextEntry) {
            if (pEntry->dwType != KFET_UNKNOWN) {
                _sntprintf(szName, countof(szName),
                  TEXT("%d: %s (ID=0x%08X%s)"),
                  (unsigned)pEntry->dwTag,
                  pEntry->szDescription,
                  (unsigned)pEntry->dwID,
                  (pEntry->dwType == KFET_PLAINTEXT) ? TEXT("") :
                  TEXT(", encrypted"));
            } else {
                _sntprintf(szName, countof(szName), TEXT("%d: (%s)"),
                  (unsigned)pEntry->dwTag, pEntry->fRandom ?
                  TEXT("random") : TEXT("unknown format"));
            }
            SendMessage(pkfdd->hwndList, LB_ADDSTRING, 0, (LPARAM)szName);
        }
        if (lCurSel == LB_ERR) {
            lCurSel = pkfdd->dwCount - 1;
        }
    } else {
        lCurSel = 0;
    }
    SendMessage(pkfdd->hwndList, LB_ADDSTRING, 0, (LPARAM)TEXT(""));
    SendMessage(pkfdd->hwndList, LB_SETCURSEL, lCurSel, 0);
}

static PKF_ENTRY GetSelectedEntry(PKF_DIALOG_DATA pkfdd,
  LRESULT *plCurSel, PKF_ENTRY *ppPrevEntry)
{
    PKF_ENTRY pEntry;
    PKF_ENTRY pPrevEntry;
    LRESULT   lCurSel;

    lCurSel = SendMessage(pkfdd->hwndList, LB_GETCURSEL, 0, 0);
    if (lCurSel == LB_ERR) {
        if (plCurSel) {
            *plCurSel = LB_ERR;
        }
        if (ppPrevEntry) {
            *ppPrevEntry = NULL;
        }
        return NULL;
    }

    if (plCurSel) {
        *plCurSel = lCurSel;
    }

    pPrevEntry = NULL;
    pEntry = pkfdd->pFirstEntry;
    while (lCurSel--) {
        pPrevEntry = pEntry;
        pEntry = pEntry->pNextEntry;
        if (pEntry == NULL) {
            if (plCurSel) {
                *plCurSel = LB_ERR;
            }
            if (ppPrevEntry) {
                *ppPrevEntry = NULL;
            }
            return NULL;
        }
    }

    if (ppPrevEntry) {
        *ppPrevEntry = pPrevEntry;
    }

    return pEntry;
}

static void NewDocument(PKF_DIALOG_DATA pkfdd)
{
    PKF_ENTRY pEntry;
    PKF_ENTRY pNextEntry;

    *pkfdd->szFileName = TEXT('\0');
    for (pEntry = pkfdd->pFirstEntry; pEntry != NULL; pEntry = pNextEntry) {
        pNextEntry = pEntry->pNextEntry;
        free(pEntry);
    }
    pkfdd->pFirstEntry = NULL;
    pkfdd->pLastEntry = NULL;
    pkfdd->dwCount = 0;
    UpdateList(pkfdd);
    UpdateCaption(pkfdd);
}

INT_PTR CALLBACK KeyFileDialogProc(HWND hwndDlg, UINT uMsg,
  WPARAM wParam, LPARAM lParam)
{
    PKF_DIALOG_DATA pkfdd;

    switch (uMsg) {
    case WM_INITDIALOG: {
        WINDOWINFO wi;

        pkfdd = (PKF_DIALOG_DATA) lParam;
        pkfdd->fModified = FALSE;
        pkfdd->pFirstEntry = NULL;
        pkfdd->pLastEntry = NULL;
        pkfdd->dwCount = 0;

        pkfdd->pPasswordCache = pkfdd->pSecData->rgPasswordCache;
        pkfdd->pkfiDecryptedData = &pkfdd->pSecData->kfiDecryptedData;
        pkfdd->pkfiKeyData = &pkfdd->pSecData->kfiKeyData;
        pkfdd->paszPassword = pkfdd->pSecData->aszPassword;
        pkfdd->pwszTempPassword = pkfdd->pSecData->wszTempPassword;

        pkfdd->fInternalCall = FALSE;

        pkfdd->hwndDlg = hwndDlg;
        pkfdd->hwndList = GetDlgItem(hwndDlg, IDC_LIST);
        pkfdd->hWaitCursor = LoadCursor(NULL, IDC_WAIT);
        pkfdd->hArrowCursor = LoadCursor(NULL, IDC_ARROW);
        SetWindowUserData(hwndDlg, pkfdd);

        FillPasswordCache(pkfdd);

        if (*pkfdd->szFileName != TEXT('\0')) {
            TCHAR szFileName[MAX_PATH];
            CopyString(szFileName, MAX_PATH, pkfdd->szFileName);
            *pkfdd->szFileName = TEXT('\0');
            OpenKeyFile(pkfdd, szFileName);
        }

        UpdateList(pkfdd);
        UpdateCaption(pkfdd);
        if (GetWindowInfo(hwndDlg, &wi)) {
            MoveWindow(pkfdd->hwndList, 0, 0,
              wi.rcClient.right - wi.rcClient.left,
              wi.rcClient.bottom - wi.rcClient.top, FALSE);
        }

        return TRUE; }

    case WM_COMMAND:
        switch (HIWORD(wParam)) {
        case LBN_DBLCLK:
            SendMessage(hwndDlg, WM_COMMAND, ID_ITEM_EDIT, 0);
            return TRUE;
        }

        pkfdd = (PKF_DIALOG_DATA)GetWindowUserData(hwndDlg);
        switch (LOWORD(wParam)) {
        case ID_FILE_NEW:
            if (SaveQuery(pkfdd)) {
                NewDocument(pkfdd);
            }
            return TRUE;

        case ID_FILE_OPEN:
            if (SaveQuery(pkfdd)) {
                TCHAR szFileName[MAX_PATH];
                *szFileName = TEXT('\0');
                if (OpenKeyFileDialog(pkfdd, szFileName)) {
                    OpenKeyFile(pkfdd, szFileName);
                }
            }
            return TRUE;

        case ID_FILE_SAVE:
            if (pkfdd->fModified && pkfdd->pFirstEntry != NULL) {
                if (*pkfdd->szFileName) {
                    SaveFile(pkfdd, pkfdd->szFileName);
                } else {
                    TCHAR szFileName[MAX_PATH];
                    *szFileName = TEXT('\0');
                    if (SaveFileDialog(pkfdd, szFileName)) {
                        SaveFile(pkfdd, szFileName);
                    }
                }
            }
            return TRUE;

        case ID_FILE_SAVEAS:
            if (pkfdd->pFirstEntry != NULL) {
                TCHAR szFileName[MAX_PATH];
                CopyString(szFileName, MAX_PATH, pkfdd->szFileName);
                if (SaveFileDialog(pkfdd, szFileName)) {
                    SaveFile(pkfdd, szFileName);
                }
            }
            return TRUE;

        case ID_FILE_EXIT:
        case IDCANCEL:
            if (SaveQuery(pkfdd)) {
                pkfdd->dwResult = LOWORD(wParam);
            }
            return TRUE;

        case IDOK:
        case ID_ITEM_EDIT: {
            PKF_ENTRY pEntry;
            LRESULT   lCurSel;
            TCHAR     szCaption[40 + MAX_LIST_TEXT_LEN + 1];

            if ((pEntry = GetSelectedEntry(pkfdd, &lCurSel, NULL)) == NULL) {
                SendMessage(hwndDlg, WM_COMMAND, ID_ITEM_NEW, 0);
                return TRUE;
            }
            if (pEntry->fRandom) {
                return TRUE;
            }

            if (pEntry->dwType == KFET_PLAINTEXT) {
                *pkfdd->paszPassword = '\0';
                pkfdd->pkfiKeyData->fValid = FALSE;
                if (KFI_Decrypt(pkfdd->pkfiDecryptedData,
                  &pEntry->kfiEncryptedData, pkfdd->pkfiKeyData) != KFIE_OK) {
                    return TRUE;
                }
            } else {
                BOOL dwType = pEntry->dwType;
                SetCursor(pkfdd->hWaitCursor);
                if (!TryCachedPasswords(pkfdd, pEntry)) {
                    PASSWORD_DIALOG_DATA pdd;
                    memset(&pdd, '\0', sizeof(pdd));
                    pdd.pszCaption = NULL;
                    pdd.uCodePage = CP_UTF8;
                    pdd.pPassword = pkfdd->paszPassword;
                    pdd.cbPasswordSize = KFI_MAX_PASSWORD;
                    pdd.pwszTempPassword = pkfdd->pwszTempPassword;
                    pdd.dwMinPasswordLength = 0;
                    pdd.dwMaxPasswordLength = KFI_MAX_PASSWORD - 1;
                    SetCursor(pkfdd->hArrowCursor);
                    if (DialogBoxParam(pkfdd->hInstance,
                      MAKEINTRESOURCE(IDD_PASSWORD), NULL, PasswordDialogProc,
                      (LPARAM)&pdd) != IDOK) {
                        return TRUE;
                    }
                    SetCursor(pkfdd->hWaitCursor);
                    if (!TryPassword(pkfdd, pEntry)) {
                        return TRUE;
                    }
                }
                SetCursor(pkfdd->hArrowCursor);
                if (dwType == KFET_UNKNOWN) {
                    return TRUE;
                }
            }

            SendMessage(pkfdd->hwndList, LB_GETTEXT, lCurSel,
              (LPARAM)(szCaption));
            AppendString(szCaption, countof(szCaption),
              TEXT(" - Keyfile Editor"));
            if (KFI_Edit(pkfdd->pkfiDecryptedData, pkfdd->paszPassword,
              KFI_MAX_PASSWORD, szCaption)) {
                pEntry->fModified = TRUE;
                CopyStringFromUnicode(pEntry->szDescription,
                  countof(pEntry->szDescription),
                  pkfdd->pkfiDecryptedData->wszDescription);
                pEntry->dwID = pkfdd->pkfiDecryptedData->dwID;
                pEntry->dwType = *pkfdd->paszPassword ? KFET_ENCRYPTED :
                  KFET_PLAINTEXT;
                if (KFI_UpdateSalt(pkfdd->pkfiDecryptedData) != KFIE_OK) {
                    return !pkfdd->fInternalCall;
                }
                if (KFI_InitEncryptionData(pkfdd->pkfiDecryptedData,
                  pkfdd->pkfiKeyData, pkfdd->paszPassword,
                  strlen(pkfdd->paszPassword)) != KFIE_OK) {
                    return !pkfdd->fInternalCall;
                }
                if (KFI_Encrypt(&pEntry->kfiEncryptedData,
                  pkfdd->pkfiDecryptedData, pkfdd->pkfiKeyData) !=  KFIE_OK) {
                    return !pkfdd->fInternalCall;
                }
                pkfdd->fModified = TRUE;
                TryPassword(pkfdd, NULL);
                UpdateList(pkfdd);
            } else {
                return !pkfdd->fInternalCall;
            }

            return TRUE; }

        case ID_ITEM_NEW: {
            PKF_ENTRY pEntry;
            PKF_ENTRY pTempEntry;
            DWORD     dwTag = 1;
            LRESULT   lCurSel;

            if (pkfdd->dwCount >= KF_MAX_ITEMS) {
                return TRUE;
            }
            lCurSel = SendMessage(pkfdd->hwndList, LB_GETCURSEL, 0, 0);
            for (pEntry = pkfdd->pFirstEntry; pEntry != NULL;
              pEntry = pEntry->pNextEntry) {
                if (pEntry->dwTag >= dwTag) {
                    dwTag = pEntry->dwTag + 1;
                }
            }
            if ((pEntry = malloc(sizeof(KF_ENTRY))) == NULL) {
                return TRUE;
            }
            memset(pEntry, '\0', sizeof(*pEntry));
            pEntry->dwType = KFET_PLAINTEXT;
            KFI_Init(pkfdd->pkfiDecryptedData);
            pEntry->dwID = pkfdd->pkfiDecryptedData->dwID;
            pTempEntry = pkfdd->pFirstEntry;
            while (pTempEntry != NULL) {
                if (pTempEntry->dwType != KFET_UNKNOWN &&
                  pTempEntry != pEntry) {
                    if (pTempEntry->dwID == pEntry->dwID) {
                        pEntry->dwID++;
                        if (pEntry->dwID == 0) {
                            pEntry->dwID = 1;
                        }
                        pTempEntry = pkfdd->pFirstEntry;
                        continue;
                    }
                }
                pTempEntry = pTempEntry->pNextEntry;
            }
            KFI_InitEncryptionData(pkfdd->pkfiDecryptedData,
              pkfdd->pkfiKeyData, NULL, 0);
            KFI_Encrypt(&pEntry->kfiEncryptedData,
              pkfdd->pkfiDecryptedData, pkfdd->pkfiKeyData);
            pEntry->dwTag = dwTag;
            if (pkfdd->dwCount) {
                pkfdd->pLastEntry->pNextEntry = pEntry;
            } else {
                pkfdd->pFirstEntry = pEntry;
            }
            pkfdd->dwCount++;
            pkfdd->pLastEntry = pEntry;
            UpdateList(pkfdd);
            SendMessage(pkfdd->hwndList, LB_SETCURSEL, pkfdd->dwCount-1, 0);
            if (lCurSel >= 0) {
                unsigned n;
                n = (unsigned)(pkfdd->dwCount - lCurSel) - 1;
                while (n--) {
                    SendMessage(hwndDlg, WM_COMMAND, ID_ITEM_MOVE_UP, 0);
                }
            }
            if (!pkfdd->fInternalCall) {
                pkfdd->fInternalCall = TRUE;
                if (!KeyFileDialogProc(hwndDlg, WM_COMMAND, ID_ITEM_EDIT, 0)) {
                    KeyFileDialogProc(hwndDlg, WM_COMMAND, ID_ITEM_DELETE, 0);
                }
                pkfdd->fInternalCall = FALSE;
            }
            return TRUE; }

        case ID_ITEM_NEW_RANDOM: {
            PKF_ENTRY pEntry;
            DWORD     dwTag = 1;

            if (pkfdd->dwCount >= KF_MAX_ITEMS) {
                return TRUE;
            }
            for (pEntry = pkfdd->pFirstEntry; pEntry != NULL;
              pEntry = pEntry->pNextEntry) {
                if (pEntry->dwTag >= dwTag) {
                    dwTag = pEntry->dwTag + 1;
                }
            }
            if ((pEntry = malloc(sizeof(KF_ENTRY))) == NULL) {
                return TRUE;
            }
            memset(pEntry, '\0', sizeof(*pEntry));
            pEntry->dwType = KFET_UNKNOWN;
            pEntry->fRandom = TRUE;
#if 0
            GetRandom(&pEntry->kfiEncryptedData, sizeof(KFI_DATA));
#else
            KFI_Init(pkfdd->pkfiDecryptedData);
            GetRandom(pkfdd->pkfiKeyData, sizeof(KFI_ENCRYPTION_DATA));
            pkfdd->pkfiKeyData->fValid = TRUE;
            KFI_Encrypt(&pEntry->kfiEncryptedData,
              pkfdd->pkfiDecryptedData, pkfdd->pkfiKeyData);
            KFI_UpdateSalt(&pEntry->kfiEncryptedData);
#endif
            pEntry->dwTag = dwTag;
            if (pkfdd->dwCount) {
                pkfdd->pLastEntry->pNextEntry = pEntry;
            } else {
                pkfdd->pFirstEntry = pEntry;
            }
            pkfdd->dwCount++;
            pkfdd->pLastEntry = pEntry;
            pkfdd->fModified = TRUE;
            UpdateList(pkfdd);
            SendMessage(pkfdd->hwndList, LB_SETCURSEL, pkfdd->dwCount-1, 0);

            if (pkfdd->dwCount > 1) {
                DWORD dwTemp;
                GetRandom(&dwTemp, sizeof(dwTemp));
                dwTemp = dwTemp % pkfdd->dwCount;
                while (dwTemp--) {
                    SendMessage(hwndDlg, WM_COMMAND, ID_ITEM_MOVE_UP, 0);
                }
            }

            return TRUE; }

        case ID_ITEM_DELETE: {
            PKF_ENTRY pEntry;
            PKF_ENTRY pPrevEntry;

            if ((pEntry = GetSelectedEntry(pkfdd, NULL, &pPrevEntry)) == NULL) {
                return TRUE;
            }
            if (pPrevEntry) {
                pPrevEntry->pNextEntry = pEntry->pNextEntry;
            } else {
                pkfdd->pFirstEntry = pEntry->pNextEntry;
            }
            if (pEntry->pNextEntry == NULL) {
                pkfdd->pLastEntry = pPrevEntry;
            }
            pkfdd->dwCount--;
            UpdateList(pkfdd);
            free(pEntry);
            pkfdd->fModified = TRUE;
            return TRUE; }

        case ID_ITEM_MOVE_UP: {
            LRESULT lCurSel;

            if (GetSelectedEntry(pkfdd, &lCurSel, NULL) == NULL) {
                return TRUE;
            }
            if (lCurSel == 0) {
                return TRUE;
            }
            SendMessage(pkfdd->hwndList, LB_SETCURSEL, lCurSel - 1, 0);
            pkfdd->fInternalCall = TRUE;
            SendMessage(hwndDlg, WM_COMMAND, ID_ITEM_MOVE_DOWN, 0);
            pkfdd->fInternalCall = FALSE;
            return TRUE; }

        case ID_ITEM_MOVE_DOWN: {
            LRESULT   lCurSel;
            PKF_ENTRY pEntry;
            PKF_ENTRY pPrevEntry;
            PKF_ENTRY pNextEntry;

            if ((pEntry = GetSelectedEntry(pkfdd, &lCurSel, &pPrevEntry)) ==
              NULL) {
                return TRUE;
            }
            if ((pNextEntry = pEntry->pNextEntry) == NULL) {
                return TRUE;
            }
            if (pPrevEntry) {
                pPrevEntry->pNextEntry = pNextEntry;
            } else {
                pkfdd->pFirstEntry = pNextEntry;
            }
            pEntry->pNextEntry = pNextEntry->pNextEntry;
            pNextEntry->pNextEntry = pEntry;
            if (pEntry->pNextEntry == NULL) {
                pkfdd->pLastEntry = pEntry;
            }
            if (!pkfdd->fInternalCall) {
                SendMessage(pkfdd->hwndList, LB_SETCURSEL, lCurSel + 1, 0);
            }
            pkfdd->fModified = TRUE;
            UpdateList(pkfdd);

            return TRUE; }
        }
        break;

    case WM_SIZE:
        pkfdd = (PKF_DIALOG_DATA)GetWindowUserData(hwndDlg);
        MoveWindow(pkfdd->hwndList, 0, 0, LOWORD(lParam), HIWORD(lParam),
          TRUE);
        return TRUE;

    case WM_QUERYENDSESSION:
        pkfdd = (PKF_DIALOG_DATA)GetWindowUserData(hwndDlg);
        return !SaveQuery(pkfdd);
    }

    return FALSE;
}

BOOL KF_Edit(PKF_DIALOG_DATA pkfdd)
{
    HWND   hwndDlg;
    HACCEL hAccTable;
    MSG    msg;
    BOOL   bResult;
    DWORD  *pdwResult;

    if (pkfdd->hInstance == NULL) {
        pkfdd->hInstance = GetModuleHandle(NULL);
    }
    if ((hwndDlg = CreateDialogParam(pkfdd->hInstance,
      MAKEINTRESOURCE(IDD_KEYFILE), NULL, KeyFileDialogProc,
      (LPARAM)pkfdd)) == NULL) {
        return FALSE;
    }
    if ((hAccTable = LoadAccelerators(pkfdd->hInstance,
      MAKEINTRESOURCE(IDR_ACCELERATORS))) == NULL) {
        DestroyWindow(hwndDlg);
        return FALSE;
    }
    ShowWindow(hwndDlg, SW_SHOWNORMAL);
    pdwResult = &pkfdd->dwResult;
    *pdwResult = 0;

    while ((bResult = GetMessage(&msg, NULL, 0, 0)) != 0) {

        if (bResult == -1) {
            /* error */
            break;
        }

        FillRandomPool(&msg, sizeof(msg));

        if (!TranslateAccelerator(hwndDlg, hAccTable, &msg)) {
            if (msg.message == WM_KEYDOWN && msg.wParam == VK_ESCAPE) {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
                continue;
            }

            if (IsDialogMessage(hwndDlg, &msg)) {
                if (*pdwResult) break;
            } else {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        }
    }
    DestroyWindow(hwndDlg);

    return TRUE;
}
